# CCND
# Certified Crime Network Database

from math import sqrt, pi, sin, asin, radians, cos, acos, atan2, tan
from random import triangular, randint, choice, random, randrange, uniform
from collections import deque
from glob import glob, iglob
from re import match
from os import listdir
from os.path import join, basename, exists
from sys import argv
from copy import copy, deepcopy

from pygame import Surface, PixelArray, Color, Rect
from pygame.font import Font
from pygame.mixer import Sound, set_num_channels
from pygame.draw import line, polygon, circle
from pygame.time import get_ticks
from pygame.image import load, save
from pygame.event import clear
from pygame.transform import rotate, scale
from pygame.gfxdraw import aacircle
from pygame.locals import *

class SoundEffect(Sound):

    def __init__(self, path, volume):
        Sound.__init__(self, path)
        self.set_volume(volume)

    def play(self, loops=0, maxtime=0, fade_ms=0, position=None):
        channel = Sound.play(self, loops, maxtime, fade_ms)
	if position is not None and channel is not None:
            right = 1 + min(0, ((position - .5) * 2))
            left = 1 - max(0, ((position - .5) * 2))
            channel.set_volume(left, right)

from seam.pgfw.Game import Game
from seam.pgfw.GameChild import GameChild
from seam.pgfw.Sprite import Sprite
from seam.pgfw.Animation import Animation
from seam.pgfw.Note import Note

class Seam(Game, Animation):

    METER_BEVEL = 15
    METER_SIZE = 300, 22
    METER_ALPHA = 255
    METER_TICK_WIDTH = 3
    METER_TICK_MARGIN = 3
    METER_SHADOW_ALPHA = 120

    def __init__(self):
        Game.__init__(self)
        Animation.__init__(self, self)
        set_num_channels(16)
        self.delegate = self.get_game().delegate
        self.capturing = False
        self.ground = Ground(self)
        self.map = Map(self)
        self.timer = Timer(self)
        self.wires = Wires(self)
        self.ships = Ships(self)
        self.net = Net(self)
        self.enemies = Enemies(self)
        self.life_meter = LifeMeter(self)
        self.grid = Grid(self)
        self.train = Train(self)
        self.city = City(self)
        self.success_message = EndMessage(self, "SUCCESS", 0)
        self.failure_message = EndMessage(self, "GAME OVER", 0)
        self.instructions = Instructions(self)
        self.high_scores = HighScores(self)
        self.next_sound_effect = SoundEffect(self.get_resource("next.wav"), .9)
        self.introduction_audio = Sound(self.get_resource("Beam.ogg"))
        self.game_audio = Sound(self.get_resource("Mouth.ogg"))
        self.end_audio = Sound(self.get_resource("Tode.ogg"))
        self.register(self.release_game_over_wait_period,
                      self.release_instructions_display_wait_period)
        self.reset()
        self.subscribe(self.respond)
        clear()

    def reset(self):
        self.game_over = False
        self.success = False
        self.game_over_wait_period_over = True
        self.displaying_instructions = True
        self.instructions_display_wait_period_over = False
        self.play(self.release_instructions_display_wait_period, delay=800,
                  play_once=True)
        self.map.reset()
        self.end_audio.fadeout(1000)
        self.game_audio.fadeout(1000)
        self.introduction_audio.play(-1, 0, 1000)
        self.ground.reset()
        self.timer.reset()
        self.ships.reset()
        self.wires.reset()
        self.net.reset()
        self.enemies.reset()
        self.life_meter.reset()
        self.grid.reset()
        self.train.reset()
        self.success_message.reset()
        self.failure_message.reset()
        self.instructions.reset()
        self.high_scores.reset()

    def release_instructions_display_wait_period(self):
        self.instructions_display_wait_period_over = True

    def respond(self, event):
        compare = self.delegate.compare
        if not self.game_over and not self.displaying_instructions:
            if compare(event, "act"):
                self.net.start()
            elif compare(event, "act", True) and self.net.active:
                self.net.capture()
        if compare(event, "reset-game"):
            self.reset()
        elif self.game_over and self.game_over_wait_period_over and \
                 compare(event, "any"):
            self.reset()
            self.next_sound_effect.play()
            self.end_audio.fadeout(1000)
        elif self.displaying_instructions and \
             self.instructions_display_wait_period_over and \
             compare(event, "any"):
            self.displaying_instructions = False
            self.instructions.deactivate()
            self.high_scores.deactivate()
            self.timer.start()
            self.enemies.spawn_initial()
            self.enemies.activate_spawning()
            self.next_sound_effect.play()
            self.introduction_audio.fadeout(1000)
            self.game_audio.play(-1, 0, 0)
            self.life_meter.show()

    def end_game(self, success):
        self.enemies.halt_spawning()
        self.ships.cancel_motion()
        self.net.deactivate()
        self.game_audio.fadeout(1000)
        self.end_audio.play(-1, 0, 1000)
        self.timer.running = False
        self.game_over = True
        self.game_over_wait_period_over = False
        self.play(self.release_game_over_wait_period, delay=1200,
                  play_once=True)
        if success:
            scores = open(self.get_resource("scores"), "a")
            string = str(self.grid.ratio)
            scores.write(string + "\n")
            scores.close()
            self.success_message.activate()
            self.high_scores.refresh(string)
            self.next_sound_effect.play()
        else:
            self.failure_message.activate()
            self.high_scores.cancel_blink()
            self.enemies.hit_sound_effect.play()
        self.success = success

    def release_game_over_wait_period(self):
        self.game_over_wait_period_over = True

    def update(self):
        self.map.update()
        self.train.update()
        self.ground.update()
        self.train.draw_hidden()
        self.map.update_expressions()
        self.wires.update()
        self.ships.shift()
        self.net.update()
        self.ships.update()
        self.enemies.update()
        if not self.game_over and self.life_meter.lives == 0:
            self.end_game(False)
        self.city.update()
        self.life_meter.update()
        self.timer.update()
        self.grid.update()
        self.failure_message.update()
        self.success_message.update()
        self.instructions.update()
        self.high_scores.update()
        Animation.update(self)


class Ground(Sprite):

    BASE_COLOR = Color(225, 192, 87)
    TRANSPARENT_COLOR = 0, 0, 0
    FRAME_COUNT = 12
    SIGN_FOREGROUND = 255, 255, 255
    FRAMERATE = 100

    def __init__(self, parent):
        Sprite.__init__(self, parent, self.FRAMERATE)
        self.display_surface = self.get_display_surface()

    def reset(self):
        self.clear_frames()
        self.set_frames()

    def set_frames(self):
        ds = self.display_surface
        background = Surface(ds.get_size())
        tile = load(self.get_resource("brick.png")).convert()
        for x in xrange(0, background.get_width(), tile.get_width()):
            for y in xrange(0, background.get_height(), tile.get_height()):
                background.blit(tile, (x, y))
        root = self.get_resource("sign")
        index = None
        for arg in argv:
            result = match("^-([\-0-9][0-9]*)$", arg)
            if result:
                index = int(result.group(1))
        paths = listdir(root)
        if index is None:
            index = randrange(0, len(paths))
        file_name = basename(sorted(paths)[index])
        image = load(join(root, file_name)).convert()
        image.set_colorkey(self.TRANSPARENT_COLOR)
        for x in xrange(image.get_width()):
            for y in xrange(image.get_height()):
                if (x + y) % 2:
                    image.set_at((x, y), self.TRANSPARENT_COLOR)
        area_r = Rect(0, 0, 700, 420)
        area_r.center = ds.get_rect().center
        image_r = image.get_rect()
        dx = (area_r.w - image_r.w) / self.FRAME_COUNT
        dy = -(area_r.h - image_r.h) / self.FRAME_COUNT
        image_r.bottomleft = area_r.bottomleft
        if dx % 2 != dy % 2:
            dx += 1
            image_r.left -= self.FRAME_COUNT / 2
        start_x, start_y = image_r.topleft
        fc = self.FRAME_COUNT
        first = randint(0, 360)
        hues = deque(first + 25 * abs(i - fc / 2) for i in xrange(fc))
        color = Color(0, 0, 0)
        for ii in xrange(fc):
            x, y = start_x, start_y
            frame = background.copy()
            frame.set_colorkey(self.TRANSPARENT_COLOR)
            for ii, hue in enumerate(hues):
                sign = image.copy()
                sign.set_alpha(int(31 + (255 - 31) * float(ii) / (fc - 1)))
                color.hsla = hue % 360, 100, 50, 100
                pixels = PixelArray(sign)
                pixels.replace(self.SIGN_FOREGROUND, color)
                del pixels
                frame.blit(sign, (int(x), int(y)))
                x += dx
                y += dy
            hues.rotate(-1)
            self.add_frame(frame)

    def update(self):
        opened = self.get_current_frame().copy()
        for rect in self.parent.map.revealed:
            opened.fill(self.TRANSPARENT_COLOR, rect)
        Sprite.update(self, substitute=opened)


class Map(Sprite):

    SHADOW_THICKNESS = 8
    SHADOW_ALPHA = 178
    HORIZONTAL, VERTICAL = range(2)
    PADDING = 120

    def __init__(self, parent):
        Sprite.__init__(self, parent, 125)
        self.set_frames()
        self.set_shadow_strips()
        self.neurons = Neurons(self)

    def set_frames(self):
        for path in sorted(iglob(join(self.get_resource("display",
                                                        "map-frame-path"),
                                     "*.png"))):
            self.add_frame(load(path).convert())

    def set_shadow_strips(self):
        tile = Surface([self.SHADOW_THICKNESS] * 2, SRCALPHA)
        tw, th = tile.get_size()
        for y in xrange(th):
            tile.fill((0, 0, 0, self.SHADOW_ALPHA * (1 - float(y) / (th - 1))),
                      (0, y, tw, 1))
        ds = self.get_display_surface()
        horizontal, vertical = self.shadow_strips = (Surface((ds.get_width() + \
                                                              self.PADDING,
                                                              th), SRCALPHA),
                                                     Surface((tw,
                                                              ds.get_height() \
                                                              + self.PADDING),
                                                             SRCALPHA))
        for x in xrange(0, horizontal.get_width(), tw):
            horizontal.blit(tile, (x, 0))
        for y in xrange(0, vertical.get_height(), tw):
            vertical.blit(rotate(tile, 90), (0, y))
        diagonal = self.diagonal = tile.copy()
        for w in xrange(tw, 0, -1):
            diagonal.fill((0, 0, 0, self.SHADOW_ALPHA * (1 - float(w) / tw)),
                          (0, 0, w, w))

    # we had a good dog sperm
    # see you in the afterlife
    # psychotic thoughts of blowing people away one by one
    # BLAT BLAT BLAT
    # everybody goes down
    # bananas I'm losing it
    # I'm going bananas
    # there's nothing but a hole
    # ...and 12 packs of crayons
    # snuffed out at 187 years old

    def reset(self):
        self.revealed = []
        self.shadow_markers = []
        self.neurons.reset()

    def reveal(self, rect):
        if rect not in self.revealed:
            self.revealed.append(rect)
            self.set_shadow_markers()
            self.neurons.check_visibility()

    def set_shadow_markers(self):
        shadow_markers = self.shadow_markers = [], []
        for rect in self.revealed:
            for orientation in self.HORIZONTAL, self.VERTICAL:
                if orientation == self.HORIZONTAL:
                    keep = (rect.left, rect.right),
                else:
                    keep = (rect.top, rect.bottom),
                unset = []
                for other in self.revealed:
                    if rect.colliderect(other) or rect.right == other.left or \
                           rect.left == other.right:
                        if orientation == self.HORIZONTAL:
                            if other.top < rect.top <= other.bottom:
                                unset.append((other.left, other.right))
                        else:
                            if other.left < rect.left <= other.right:
                                unset.append((other.top, other.bottom))
                for other in unset:
                    r_keep = []
                    for low, high in keep:
                        if other[0] > high or other[1] < low:
                            r_keep.append((low, high))
                        else:
                            under = low, other[0]
                            over = other[1], high
                            if under[0] < under[1]:
                                r_keep.append(under)
                            if over[0] < over[1]:
                                r_keep.append(over)
                    keep = r_keep
                shadow_markers[orientation].append((rect, keep))

    def update(self):
        Sprite.update(self, self.revealed)
        self.neurons.update()
        ds = self.display_surface
        for ii, markers in enumerate(self.shadow_markers):
            strip = self.shadow_strips[ii]
            for rect, keep in markers:
                for low, high in keep:
                    if ii == self.HORIZONTAL:
                        position = low, rect.top
                        w, h = high - low, strip.get_height()
                    else:
                        position = rect.left, low
                        w, h = strip.get_width(), high - low
                    ds.blit(strip, position, (0, 0, w, h))
                    # ds.blit(strip, position)
                    if ii == self.HORIZONTAL:
                        ds.blit(self.diagonal, (high, rect.top))

    def update_expressions(self):
        self.neurons.update_expressions()


class Neurons(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.root_directory = self.get_resource("display", "neuron-path")
        self.set_shadow()
        self.set_neurons()
        self.expression_surface = load(join(self.root_directory,
                                            "joy.png")).convert_alpha()
        self.found_sound_effect = SoundEffect(self.get_resource("lick.wav"), .5)

    def reset(self):
        self.active_expressions = []
        for neuron in self.neurons:
            neuron.reset()

    def set_shadow(self):
        shadow = self.shadow = Sprite(self)
        shadow.load_from_path(join(self.root_directory, "shadow.png"), True)

    def set_neurons(self):
        neurons = self.neurons = []
        for ii, line in enumerate(open(join(self.root_directory, "map"))):
            values = tuple(int(x) for x in line.split())
            if ii == 0:
                self.shadow.location.topleft = values[2:4]
            else:
                self.shadow.add_location(values[2:4])
            neurons.append(Neuron(self, ii, values[0:2]))

    def check_visibility(self):
        for neuron in (n for n in self.neurons if not n.revealed):
            area = None
            for rect in self.parent.revealed:
                if rect.colliderect(neuron.location):
                    piece = rect.clip(neuron.location)
                    if not area:
                        area = piece
                    else:
                        area.union_ip(piece)
                    if area and area.w >= neuron.location.w and \
                           area.h >= neuron.location.h:
                        neuron.revealed = True
                        self.active_expressions.append(Expression(
                            self, neuron.location.topleft))
                        self.found_sound_effect.play(position=neuron.
                                                     location.centerx)
                        break

    def update(self):
        self.shadow.update()
        for neuron in self.neurons:
            neuron.update()

    def update_expressions(self):
        for expression in self.active_expressions:
            expression.update()


class Neuron(Sprite):

    def __init__(self, parent, index, coordinates):
        Sprite.__init__(self, parent, 240)
        self.load_from_path(join(self.parent.root_directory, str(index)),
                            True)
        self.location.topleft = coordinates

    def reset(self):
        self.revealed = False


class Expression(Sprite):

    def __init__(self, parent, base):
        Sprite.__init__(self, parent)
        self.add_frame(self.parent.expression_surface)
        self.location.topleft = base[0] + 15, base[1] - 13
        self.register(self.disappear)
        self.play(self.disappear, play_once=True, delay=2500)

    def disappear(self):
        self.parent.active_expressions.remove(self)


class Timer(GameChild):

    LIMIT = 99999
    SIZE = 31, 40
    COLOR = 44, 106, 255
    TRANSPARENT_COLOR = 255, 0, 255

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.font = Font(self.get_resource("display", "timer-font"), 34)

    def reset(self):
        self.elapsed = 0
        self.running = False
        self.last_ticks = None
        self.set_tick_surface()

    def set_tick_surface(self):
        surface = self.tick_surface = Surface(self.parent.METER_SIZE)
        surface.fill(self.COLOR)
        surface.set_colorkey(self.TRANSPARENT_COLOR)
        rect = surface.get_rect()
        tw, tm = self.parent.METER_TICK_WIDTH, self.parent.METER_TICK_MARGIN
        for x in xrange(tw, rect.w, tm + tw):
            surface.fill(self.TRANSPARENT_COLOR, (x, 0, tw, rect.h))
        bw = abs(rect.h / tan(radians(self.parent.METER_BEVEL)))
        polygon(surface, self.TRANSPARENT_COLOR, ((rect.w, 0), (rect.w, rect.h),
                                                  (rect.w - bw, rect.h)))
        surface.set_alpha(self.parent.METER_ALPHA)
        self.set_shadow_surface()

    def set_shadow_surface(self):
        color = Color(*self.COLOR)
        h, s, l, a = color.hsla
        color.hsla = h, s - 15, l - 20, a
        surface = self.shadow_surface = self.tick_surface.copy()
        pixels = PixelArray(surface)
        pixels.replace(self.COLOR, color)
        del pixels
        surface.set_alpha(self.parent.METER_SHADOW_ALPHA)

    def start(self):
        self.running = True
        self.last_ticks = None

    def update(self):
        ticks = get_ticks()
        if self.running and self.last_ticks is not None:
            self.elapsed += ticks - self.last_ticks
            if self.elapsed >= self.LIMIT:
                self.elapsed = self.LIMIT
                self.parent.end_game(True)
        self.last_ticks = ticks
        remaining = str(int(self.LIMIT - self.elapsed) / 1000)
        self.mask()
        self.display_surface.blit(self.shadow_surface, (-1, 1))
        self.display_surface.blit(self.tick_surface, (0, 0))

    def mask(self):
        tr = self.tick_surface.get_rect()
        width = int(float(self.elapsed) / self.LIMIT * tr.w)
        for surface in self.tick_surface, self.shadow_surface:
            surface.fill(self.TRANSPARENT_COLOR, (tr.w - width, 0, width, tr.h))


class Wires(GameChild, list):

    AMPLITUDES = 8, 10, 13, 16, 18, 16, 13, 10
    WIDTH = 9592
    LEFT, RIGHT = -1, 1
    SPEEDS = 1, 2, 8
    SLOW, NORMAL, FAST = range(3)

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        list.__init__(self, (Wire(self, self.LEFT), Wire(self, self.RIGHT)))

    def reset(self):
        self.set_speed()
        for wire in self:
            wire.reset()

    def set_speed(self, speed=NORMAL):
        self.speed = self.SPEEDS[speed]

    def update(self):
        for wire in self:
            wire.update()


class Wire(Sprite):

    TRANSPARENT_COLOR = 255, 0, 255

    def __init__(self, parent, heading):
        Sprite.__init__(self, parent)
        self.heading = heading
        self.display_surface = self.get_display_surface()
        self.color = (255, 255, 255) if heading == parent.LEFT else \
                     (108, 108, 108)
        self.reset()
        self.add_location(offset=(self.location.w, 0))
        self.register(self.increment_amplitude_index, interval=100)
        self.play(self.increment_amplitude_index)

    def reset(self):
        self.amplitude_index = 0
        self.set_points()

    def set_points(self):
        points = self.points = []
        next_jump = 0
        x = randint(-90, 90)
        while x < self.parent.WIDTH:
            if next_jump == 0 and x < self.parent.WIDTH - 800:
                random_point = self.get_random_point(x)
                while len(points) > 0 and \
                          abs(random_point[1] - points[-1][1]) < 80:
                    random_point = self.get_random_point(x)
                points.append(random_point)
                next_jump = int(round(triangular(2, 5)))
            x += randint(218, 402)
            points.append((x, self.get_random_y(points[-1][1])))
            next_jump -= 1

    def get_random_point(self, x):
        return x, self.get_random_y()

    def get_random_y(self, previous=None):
        if previous is None:
            previous = self.display_surface.get_rect().centery
        deviation = 73
        y = previous + abs(deviation - triangular(-deviation, deviation)) * \
            choice((1, -1))
        upper, lower = 12, self.display_surface.get_height() - 12
        if y < upper:
            y += (upper - y) * 2
        elif y > lower:
            y -= (y - lower) * 2
        return y

    def get_y(self, x, amplitude=None):
        if amplitude is None:
            amplitude = self.get_current_amplitude()
        while x < 0:
            x += self.parent.WIDTH
        x %= self.parent.WIDTH
        index = 0
        points = self.points
        while x > points[index + 1][0]:
            index += 1
        px, py = points[index]
        nx, ny = points[index + 1]
        dx = nx - px
        distance = sqrt(dx ** 2 + (ny - py) ** 2)
        h = int(round(distance * float(x - px) / dx))
        cycles = 4 * pi / distance
        beta = radians(90) - asin((nx - px) / distance)
        if ny < py:
            beta *= -1
        sx = amplitude * sin(cycles * h) + py
        return sin(beta) * h - cos(beta) * (sx - py) + py

    def get_current_amplitude(self):
        return self.parent.AMPLITUDES[self.amplitude_index]

    def increment_amplitude_index(self):
        self.amplitude_index += 1
        if self.amplitude_index >= len(self.parent.AMPLITUDES):
            self.amplitude_index = 0

    def update(self):
        self.move(self.heading * self.parent.speed)
        if self.location.right < 0:
            self.move(self.location.width)
        if self.location.left > 0:
            self.move(-self.location.width)
        step = 10
        offset = -self.location.left
        px, py = offset, self.get_y(offset)
        ds = self.display_surface
        for x in xrange(offset, offset + ds.get_width() + step, step):
            y = self.get_y(x)
            if abs(y - py) < 16:
                line(ds, self.color, (px - offset, py), (x - offset, y))
            px, py = x, y
        Sprite.update(self)


class Ships(GameChild, list):

    INVINCIBILITY_LENGTH = 3000

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        list.__init__(self, [Ship(self, Ship.OUTER), Ship(self, Ship.INNER)])

    def reset(self):
        for ship in self:
            ship.reset()

    def shift(self):
        for ship in self:
            ship.shift()

    def cancel_motion(self):
        for ship in self:
            ship.cancel_motion()

    def set_invincible(self, length=INVINCIBILITY_LENGTH):
        for ship in self:
            ship.set_invincible(length)

    def are_invincible(self):
        return self[0].is_playing(self[0].blink)

    def update(self):
        for ship in self:
            ship.update()


class Ship(Sprite):

    OUTER, INNER = range(2)
    LEFT, RIGHT = range(2)
    BLINK_INTERVAL = 150

    def __init__(self, parent, position):
        Sprite.__init__(self, parent, 240)
        self.position = position
        self.wire = parent.parent.wires[0] if \
                    position == self.OUTER else parent.parent.wires[1]
        self.display_surface = self.get_display_surface()
        self.delegate = self.get_delegate()
        self.set_frames()
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.register(self.cancel_blink)
        self.subscribe(self.respond)
        self.cancel_motion()

    def blink(self):
        self.toggle_hidden()

    def cancel_blink(self):
        self.halt(self.blink)
        self.unhide()

    def set_invincible(self, length):
        self.play(self.blink)
        self.play(self.cancel_blink, delay=length, play_once=True)

    def cancel_motion(self):
        self.moving = [False, False]

    def reset(self):
        self.remove_locations()
        self.location.centerx = self.display_surface.get_rect().centerx
        self.add_location(offset=(self.display_surface.get_width(), 0))
        self.cancel_blink()

    def set_frames(self):
        max_width = 18
        hue = 200 if self.position == self.INNER else 320
        foreground = Color(0, 0, 0)
        background = Color(0, 0, 0)
        for ii, width in enumerate(xrange(max_width, 10, -2)):
            frame = Surface((max_width, max_width))
            frame.set_colorkey((0, 0, 0))
            fr = frame.get_rect()
            rect = Rect((0, 0, width - 2, width - 2))
            rect.center = fr.center
            foreground.hsla = hue, 100, 85, 100
            background.hsla = hue, 100, 30, 100
            polygon(frame, background, ((rect.left, rect.top),
                                        (rect.right, rect.top),
                                        (rect.right, rect.bottom),
                                        (rect.left, rect.bottom)), 1)
            rect.move_ip(-1, -1)
            polygon(frame, foreground, ((rect.left, rect.top),
                                        (rect.right, rect.top),
                                        (rect.right, rect.bottom),
                                        (rect.left, rect.bottom)), 1)
            reticule = (255, 0, 0) if self.position == self.INNER else \
                       (255, 255, 0)
            if ii % 2 == 1:
                polygon(frame, reticule, ((fr.centerx - 2, fr.centery - 2),
                                          (fr.centerx + 1, fr.centery - 2),
                                          (fr.centerx + 1, fr.centery + 1),
                                          (fr.centerx - 2, fr.centery + 1)))
            hue = (hue - 16) % 360
            self.add_frame(frame)

    def respond(self, event):
        if not self.parent.parent.game_over and not \
               self.parent.parent.displaying_instructions:
            compare = self.delegate.compare
            if compare(event, "right") or compare(event, "right", True):
                self.moving[self.RIGHT] = not event.cancel
            elif compare(event, "left") or compare(event, "left", True):
                self.moving[self.LEFT] = not event.cancel

    def shift(self):
        speed = 6
        if self.moving[self.LEFT]:
            self.move(-speed)
        if self.moving[self.RIGHT]:
            self.move(speed)
        bound = self.display_surface.get_width()
        if self.location.right < 0:
            self.move(bound)
        if self.location.left > bound:
            self.move(-bound)
        for location in self.locations:
            location.centery = self.wire.get_y(self.location.centerx - \
                                               self.wire.location.left)


class Net(GameChild):

    GROWTH_RATES = .375, .8

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        self.hues = deque(xrange(0, 360, 40))
        self.growing_sound_effect = SoundEffect(self.get_resource("grow.wav"),
                                                1)
        self.capture_sound_effect = SoundEffect(self.\
                                    get_resource("capture.wav"), 1)

    def reset(self):
        self.deactivate()
        self.init_rect()
        self.set_growth_rate()

    def deactivate(self):
        self.active = False
        self.growing_sound_effect.stop()

    def init_rect(self):
        rect = self.rect = Rect(0, 0, 0, 0)
        return rect

    def set_growth_rate(self, quick=False):
        self.growth_rate = self.GROWTH_RATES[quick]

    def start(self):
        self.ox, self.oy = 19, 19
        self.activate()

    def activate(self):
        self.active = True
        self.growing_sound_effect.play(-1)

    def capture(self):
        self.deactivate()
        for rect in self.drawn_rects:
            territory = rect.clip(self.display_surface.get_rect())
            self.parent.map.reveal(rect)
            self.parent.grid.add(territory)
        self.capture_sound_effect.play()

    def collide(self, other):
        for rect in self.drawn_rects:
            if rect.colliderect(other):
                return True

    def update(self):
        if self.active:
            self.ox += self.growth_rate
            self.oy += self.growth_rate
            rect = self.init_rect()
            ships = self.parent.ships
            rect.width = int(self.ox) * 2
            distance = abs(ships[0].location.centery - \
                           ships[1].location.centery)
            rect.height = distance + int(self.oy) * 2
            rect.centerx = ships[0].location.centerx
            if ships[0].location.centery > ships[1].location.centery:
                rect.centery = ships[1].location.centery + (distance / 2)
            else:
                rect.centery = ships[0].location.centery + (distance / 2)
            color = Color(0, 0, 0)
            color.hsla = self.hues[0], 90, 70, 100
            self.draw_outlines(rect, color)
            self.hues.rotate(-1)
            color.hsla = self.hues[0], 90, 90, 100
            self.draw_outlines(rect.inflate(-4, -4), color)

    def draw_outlines(self, rect, color):
        self.drawn_rects = []
        self.draw_outline(rect, color)
        offsets = []
        dsr = self.display_surface.get_rect()
        if rect.left < 0:
            self.draw_outline(rect.move(dsr.w, 0), color)
            if rect.top < 0:
                self.draw_outline(rect.move(dsr.w, dsr.h), color)
            elif rect.bottom > dsr.h:
                self.draw_outline(rect.move(dsr.w, -dsr.h), color)
        elif rect.right > dsr.w:
            self.draw_outline(rect.move(-dsr.w, 0), color)
            if rect.top < 0:
                self.draw_outline(rect.move(-dsr.w, dsr.h), color)
            elif rect.bottom > dsr.h:
                self.draw_outline(rect.move(-dsr.w, -dsr.h), color)
        if rect.top < 0:
            self.draw_outline(rect.move(0, dsr.h), color)
        elif rect.bottom > dsr.h:
            self.draw_outline(rect.move(0, -dsr.h), color)

    def draw_outline(self, rect, color):
        self.drawn_rects.append(rect)
        polygon(self.display_surface, color,
                (rect.topleft, rect.topright, rect.bottomright,
                 rect.bottomleft), 1)


class Enemies(Animation):

    SPAWN_RATE = .0008
    INITIAL_POPULATION = 5
    SPEEDS = 1, 4, 10

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.hit_sound_effect = SoundEffect(self.get_resource("hit.wav"), 1)
        self.spawn_sound_effect = SoundEffect(self.get_resource("spawn.wav"),
                                              .9)
        self.eat_sound_effect = SoundEffect(self.get_resource("eat-enemy.wav"),
                                            .9)
        self.register(self.spawn)

    def reset(self):
        self.enemies = []
        self.halt_spawning()
        self.halt(self.spawn)
        self.current_speed_index = 1
        self.set_speed()

    def halt_spawning(self):
        self.spawning = False

    def set_speed(self, index=None):
        if index is None:
            index = self.current_speed_index
        else:
            self.current_speed_index = index
        for enemy in self.enemies:
            enemy.speed = self.SPEEDS[index]

    def spawn_initial(self):
        self.play(self.spawn, play_once=True, initial=True)

    def activate_spawning(self):
        self.spawning = True

    def spawn(self, initial=False):
        self.enemies.append(Enemy(self))
        self.spawn_sound_effect.play(\
            position=float(self.enemies[-1].location.centerx) / \
            self.display_surface.get_width())
        self.enemies[-1].materialize()
        if initial and len(self.enemies) < self.INITIAL_POPULATION:
            self.play(self.spawn, delay=1000, play_once=True, initial=True)
        self.set_speed()

    def update(self):
        Animation.update(self)
        if self.spawning and random() < self.SPAWN_RATE:
            self.spawn()
        ships = self.parent.ships
        if not ships.are_invincible() and not self.parent.game_over:
            if self.parent.net.active:
                for enemy in self.enemies:
                    if not enemy.is_materializing() and not enemy.is_edible():
                        for location in enemy.locations:
                            if self.parent.net.collide(location):
                                self.strike_ships(enemy)
                                break
                    if ships.are_invincible():
                        self.parent.net.deactivate()
                        break
            if not ships.are_invincible():
                for enemy in self.enemies:
                    if not enemy.is_materializing():
                        for location in enemy.locations:
                            if location.collidelist([ship.location for ship in \
                                                     ships]) != -1:
                                if not enemy.is_edible():
                                    self.strike_ships(enemy)
                                else:
                                    power_ups = self.parent.train.power_ups
                                    power_ups.active_power_up.cancel_eat_enemy()
                                    self.enemies.remove(enemy)
                                    self.eat_sound_effect.play()
                                break
                    if ships.are_invincible():
                        break
        for enemy in self.enemies:
            enemy.update()

    def strike_ships(self, enemy):
        self.enemies.remove(enemy)
        self.parent.life_meter.remove()
        self.hit_sound_effect.play()
        self.parent.ships.set_invincible()


class Enemy(Sprite):

    MOVE_CHANCE = .01
    BLINK_INTERVAL = 350
    MATERIALIZATION_LENGTH = 3000

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.set_frames()
        self.place()
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.register(self.cancel_blink)

    def set_frames(self):
        w = 21
        so = 2
        color = Color(0, 0, 0)
        for hue in 0, 10, 20, 30, 40, 50, 60, 50, 40, 30, 20, 10:
            frame = Surface((w + so, w + so))
            frame.set_colorkey((0, 0, 0))
            color.hsla = 0, 0, 70, 100
            polygon(frame, color, ((so, w - 1 + so), (w / 2 + so, so),
                                   (w - 1 + so, w - 1 + so)))
            for x in xrange(1 + so, w + so, so):
                frame.fill((0, 0, 0), (x, so, 1, w))
            color.hsla = hue, 100, 50, 100
            polygon(frame, color, ((0, w - 1), (w / 2, 0), (w - 1, w - 1)))
            for x in xrange(1, w, 2):
                frame.fill((0, 0, 0), (x, 0, 1, w))
            self.add_frame(frame)
        self.add_frameset(name="edible", switch=True)
        for hue in 240, 230, 220, 210, 200, 190, 180, 190, 200, 210, 220, 230:
            frame = Surface((w + so, w + so))
            frame.set_colorkey((0, 0, 0))
            color.hsla = 0, 0, 70, 100
            polygon(frame, color, ((so, so), (w + so, so),
                                   (w / 2 - 1 + so, w - 1 + so)))
            for x in xrange(1 + so, w + so, so):
                frame.fill((0, 0, 0), (x, so, 1, w))
            color.hsla = hue, 100, 50, 100
            polygon(frame, color, ((0, 0), (w, 0), (w / 2, w - 1)))
            for x in xrange(1, w, 2):
                frame.fill((0, 0, 0), (x, 0, 1, w))
            self.add_frame(frame)
        self.set_frameset(0)

    def place(self):
        self.location.center = self.get_random_location()
        while self.location.collidelist([ship.location for ship in \
                                         self.parent.parent.ships]) != -1:
            self.location.center = self.get_random_location()
        self.course = None
        ds = self.display_surface
        self.add_location(offset=(ds.get_width(), 0))
        self.add_location(offset=(0, ds.get_height()))

    def get_random_location(self):
        return randint(0, self.display_surface.get_width()), \
               randint(0, self.display_surface.get_height())

    def blink(self):
        self.toggle_hidden()

    def cancel_blink(self):
        self.halt(self.blink)
        self.unhide()

    def materialize(self):
        self.play(self.blink)
        self.play(self.cancel_blink, delay=self.MATERIALIZATION_LENGTH,
                  play_once=True)

    def set_course(self):
        self.course = [randint(21, 395), radians(randrange(0, 360))]

    def is_materializing(self):
        return self.is_playing(self.blink)

    def is_edible(self):
        return self.get_current_frameset().name == "edible"

    def update(self):
        if self.course:
            if self.course[0] < self.speed:
                self.course = None
            else:
                self.move(sin(self.course[1]) * self.speed,
                          cos(-self.course[1]) * self.speed)
                self.course[0] -= self.speed
            rect = self.display_surface.get_rect()
            if self.location.right < 0:
                self.move(rect.w)
            if self.location.left > rect.w:
                self.move(-rect.w)
            if self.location.bottom < 0:
                self.move(dy=rect.h)
            if self.location.top > rect.h:
                self.move(dy=-rect.h)
        if not self.is_materializing() and self.course is None and \
               random() < self.MOVE_CHANCE:
            self.set_course()
        Sprite.update(self)


class LifeMeter(Animation):

    INITIAL_LIVES = 3
    MARGIN = 48
    DISPLAY_LENGTH = 2000

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.heart = Heart(self)
        self.register(self.cancel_show)

    def reset(self):
        self.lives = self.INITIAL_LIVES
        self.rearrange()
        self.heart.hide()
        self.halt()

    def remove(self):
        self.lives -= 1
        if self.lives:
            self.rearrange()
            self.show()

    def add(self):
        self.lives += 1
        self.rearrange()
        self.show()

    def rearrange(self):
        sr = self.display_surface.get_rect()
        hr = self.heart.location
        count = self.lives
        self.heart.remove_locations()
        x = sr.w / 2 - (hr.w / 2) * (count - 1) - self.MARGIN / 2 * (count - 1)
        y = sr.h / 2
        hr.center = x, y
        for ii in xrange(1, count):
            self.heart.add_location(offset=((hr.w + self.MARGIN) * ii, 0))

    def show(self):
        self.heart.unhide()
        self.play(self.cancel_show, delay=self.DISPLAY_LENGTH, play_once=True)

    def blink(self):
        self.heart.toggle_hidden()

    def cancel_show(self):
        self.heart.hide()

    def update(self):
        Animation.update(self)
        self.heart.update()


class Heart(Sprite):

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.set_frames()

    def set_frames(self):
        for ii, path in enumerate(sorted(glob(join(self.get_resource("heart"),
                                                   "[0-4]*.png")))):
            self.add_frame(load(path).convert_alpha())
        self.add_frameset([0, 1, 4, 2, 0, 3, 4],
                          [300, 0, 400, 0, 400, 0, 300], switch=True)


class Grid(GameChild, list):

    TRANSPARENT_COLOR = 0, 0, 0

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        background = self.background = Surface((80, 14))
        rect = self.background_rect = background.get_rect()
        rect.midbottom = ds.get_rect().midbottom
        self.font = Font(self.get_resource("display", "completion-font"), 12)
        self.font.set_bold(True)

    def reset(self):
        ds = self.display_surface
        list.__init__(self, [[False for y in xrange(ds.get_height())] for x in \
                             xrange(ds.get_width())])
        self.ratio = 0.0
        self.set_progress_surface()

    def set_progress_surface(self):
        surface = self.progress_surface = Surface(self.parent.METER_SIZE)
        surface.set_colorkey(self.TRANSPARENT_COLOR)
        rect = self.progress_rect = surface.get_rect()
        rect.topright = self.display_surface.get_rect().topright
        color = Color(0, 0, 0)
        color.hsla = int(self.ratio * 120), 100, 55, 100
        for x in xrange(rect.w):
            surface.fill(color, (x, 0, 1, rect.h))
        tw, tm = self.parent.METER_TICK_WIDTH, self.parent.METER_TICK_MARGIN
        for x in xrange(rect.w, 0, -(tm + tw)):
            surface.fill(self.TRANSPARENT_COLOR, (x, 0, tw, rect.h))
        bw = abs(rect.h / tan(radians(self.parent.METER_BEVEL)))
        polygon(surface, self.TRANSPARENT_COLOR, ((0, 0), (0, rect.h),
                                                  (bw, rect.h)))
        surface.set_alpha(self.parent.METER_ALPHA)
        self.set_shadow_surface(color)

    def set_shadow_surface(self, color):
        surface = self.shadow_surface = self.progress_surface.copy()
        shadow_color = color.hsla[0], 92, 50, 100
        pixels = PixelArray(surface)
        pixels.replace(color, shadow_color)
        del pixels
        surface.set_alpha(self.parent.METER_SHADOW_ALPHA)

    def add(self, rect):
        for x in xrange(rect.left, rect.right):
            self[x][rect.top:rect.bottom] = [True] * (rect.bottom - rect.top)
        count = 0
        for x in self.parent.grid:
            for y in x:
                count += y
        self.ratio = float(count) / (len(self) * len(self[0]))
        self.set_progress_surface()
        self.mask()

    def mask(self):
        tr = self.progress_surface.get_rect()
        width = int(self.ratio * tr.w)
        for surface in self.progress_surface, self.shadow_surface:
            surface.fill(self.TRANSPARENT_COLOR, (0, 0, width, tr.h))

    def update(self):
        self.display_surface.blit(self.shadow_surface,
                                  self.progress_rect.move(1, 1))
        self.display_surface.blit(self.progress_surface, self.progress_rect)


class Train(Animation):

    COUNT = 7
    SPACING = 42
    SPEED = 4
    PEEK_DEPTH = 2

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.hands = tuple(Hand(self) for _ in xrange(self.COUNT))
        self.pin = Pin(self)
        self.power_ups = PowerUps(self)
        self.trap_sound_effect = SoundEffect(\
            self.get_resource(join("power-up", "hand", "wind-laser.wav")), .8)

    def reset(self):
        self.halt()
        self.escaping = False
        self.trapped = False
        hands = self.hands
        absolute = self.absolute_locations = [[0, 0]] * len(hands)
        for hand in hands:
            hand.reset()
        ds = self.display_surface
        absolute[0] = hands[0].location.center = [randrange(0, ds.get_width()),
                                                  randrange(0, ds.get_height())]
        self.direction = randrange(0, 360)
        backward = (self.direction + 180) % 360
        for ii, hand in enumerate(hands[1:]):
            backward += triangular(-40, 40)
            x = hands[ii].location.centerx + sin(radians(backward)) * \
                self.SPACING
            y = hands[ii].location.centery - cos(radians(backward)) * \
                self.SPACING
            absolute[ii + 1] = hand.location.center = [x, y]
        for hand in hands:
            hand.add_location(offset=(ds.get_width(), 0))
            hand.add_location(offset=(0, ds.get_height()))
        self.active_notes = []
        self.power_ups.reset()

    def update(self):
        Animation.update(self)
        hands = self.hands
        if not self.trapped:
            self.trapped = True
            escape = True
            for ii, hand in enumerate(self.hands):
                contained = False
                area = None
                for rect in self.parent.map.revealed:
                    if rect.colliderect(hand.location):
                        piece = rect.clip(hand.location)
                        if not area:
                            area = piece
                        else:
                            area.union_ip(piece)
                        if area and area.w >= hand.location.w and \
                               area.h >= hand.location.h:
                            contained = True
                            break
                if not contained:
                    if ii < self.PEEK_DEPTH:
                        escape = False
                    if self.escaping and ii == 0:
                        self.escaping = False
                    self.trapped = False
                    break
            if escape and not self.escaping:
                self.escaping = True
                self.direction = (self.direction + 180 + \
                                  triangular(-20, 20)) % 360
            if self.trapped:
                self.trap_sound_effect.play()
                notes = self.active_notes = []
                hue_offset = randrange(0, 360)
                for ii, hand in enumerate(reversed(self.hands)):
                    notes.append(Eighth(self, self.hands.index(hand) == 0,
                                        hue_offset))
                    notes[-1].location.center = hand.location.center
                    notes[-1].play(notes[-1].start, delay=300 * ii,
                                   play_once=True)
                    hand.set_frameset("giving")
                    hand.fade(4000, True)
                    hue_offset += 25
            else:
                absolute = self.absolute_locations
                for ii in xrange(len(hands) - 1, 0, -1):
                    x2, y2, x1, y1 = absolute[ii - 1] + absolute[ii]
                    distance = sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
                    angle = atan2(x2 - x1, y2 - y1)
                    dx, dy = sin(angle) * distance * .1, \
                             cos(angle) * distance * .1
                    hands[ii].move(dx, dy)
                    absolute[ii][0] += dx
                    absolute[ii][1] += dy
                if random() < .0625:
                    self.direction += randint(-65, 65)
                dx, dy = sin(radians(self.direction)) * self.SPEED, \
                         -cos(radians(self.direction)) * self.SPEED
                hands[0].move(dx, dy)
                absolute[0][0] += dx
                absolute[0][1] += dy
        for hand in hands:
            hand.update()
        for note in self.active_notes:
            note.update()
        self.power_ups.update()

    def draw_hidden(self):
        self.pin.remove_locations()
        draw = False
        first = True
        for hand in self.hands:
            place_pin = True
            for rect in self.parent.map.revealed:
                if rect.colliderect(hand):
                    place_pin = False
                    break
            if place_pin:
                draw = True
                if first:
                    location = self.pin.location
                    first = False
                else:
                    location = self.pin.add_location((0, 0))
                location.center = hand.location.center
        if draw:
            self.pin.update()


class Hand(Sprite):

    TRANSPARENT_COLOR = (255, 0, 255)
    FRAME_COUNT = 8

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        self.set_frames()

    def set_frames(self):
        count = self.FRAME_COUNT
        root = self.get_resource(join("power-up", "hand"))
        for state in ("open", "closed"):
            base = load(join(root, "%s.png" % state)).convert()
            base.set_colorkey(self.TRANSPARENT_COLOR)
            for _ in xrange(count):
                frame = base.copy()
                for x in xrange(frame.get_width()):
                    for y in xrange(frame.get_height()):
                        if frame.get_at((x, y)) == (0, 0, 255):
                            color = Color(0, 0, 0)
                            color.hsla = randrange(0, 360), 100, 36, 100
                            frame.set_at((x, y), color)
                self.add_frame(frame)
        self.add_frameset(range(count), 450, "open")
        self.add_frameset([ii + (ii % 2) * count for ii in range(count)], 450,
                          "giving")

    def reset(self):
        self.remove_locations()
        self.set_frameset("open")
        self.location.fader.reset()
        self.set_alpha(255)

    def update(self):
        rect = self.display_surface.get_rect()
        if self.location.right < 0:
            self.move(rect.w)
        if self.location.left > rect.w:
            self.move(-rect.w)
        if self.location.bottom < 0:
            self.move(dy=rect.h)
        if self.location.top > rect.h:
            self.move(dy=-rect.h)
        Sprite.update(self)


class Pin(Sprite):

    SIZE = 16, 16
    FRAME_COUNT = 10
    RING_COUNT = 8

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.set_frames()

    def set_frames(self):
        w, h = self.SIZE
        for frame_ii in xrange(self.FRAME_COUNT):
            lightness = 50 + 27 * (float(frame_ii + 1) / self.FRAME_COUNT)
            frame = Surface((w, h), SRCALPHA)
            frame.fill((255, 255, 255, 0))
            for circle_ii in xrange(self.RING_COUNT):
                color = Color(0, 0, 0)
                hue = int(360 * (float(circle_ii + frame_ii) / \
                                 self.RING_COUNT)) % 360
                alpha = int(100 * (float(circle_ii + 1) / self.RING_COUNT))
                color.hsla = hue, 50, lightness, alpha
                radius = int(w * (1 - float(circle_ii) / self.RING_COUNT) / 2)
                circle(frame, color, (w / 2, h / 2), radius)
            # color.hsla = 60, 100, 20, 60
            # circle(frame, color, (w / 2, h / 2), w / 2 - 3, 1)
            color.hsla = 0, 0, 30, 68
            aacircle(frame, w / 2, h / 2, w / 2 - 3, color)
            self.add_frame(frame)


class Eighth(Sprite):

    TRANSPARENT_COLOR = 255, 0, 255
    SIXTEENTH_NOTE_LENGTH = 100

    def __init__(self, parent, is_head, hue_offset):
        Sprite.__init__(self, parent)
        self.is_head = is_head
        self.hue_offset = hue_offset
        self.set_frames()
        self.register(self.cancel_explosion, self.show_explosion, self.start)
        self.hide()

    def set_frames(self):
        base = load(self.get_resource(join("power-up", "hand",
                                           "note.png"))).convert()
        base.set_colorkey(self.TRANSPARENT_COLOR)
        for hue in ((self.hue_offset + h) % 360 for h in xrange(0, 50, 5)):
            frame = base.copy()
            color = Color(0, 0, 0)
            color.hsla = hue, 95, 55, 100
            pixels = PixelArray(frame)
            pixels.replace((0, 0, 0), color)
            pixels.replace((255, 255, 255), (63, 191, 127))
            del pixels
            self.add_frame(frame)
        self.add_frameset(name="explode", switch=True)
        base = Surface((32, 32), SRCALPHA)
        rect = base.get_rect()
        for ii in xrange(15):
            frame = base.copy()
            color = Color(0, 0, 0)
            color.hsla = ii * 2, 100, 50, 100 - (ii * 2)
            circle(frame, color, rect.center, ii + 1, 1)
            self.add_frame(frame)
        self.set_frameset(0)

    def start(self):
        self.unhide()
        self.play(self.show_explosion, delay=3000, play_once=True)

    def show_explosion(self):
        self.set_frameset("explode")
        self.play(self.cancel_explosion, delay=540, play_once=True)
        for _ in xrange(3):
            shape = choice((Note.SQUARE, Note.TRIANGLE, Note.SINE, Note.SAW,
                            Note.DIRTY))
            Note(choice(Note.names), randint(2, 5), shape=shape,
                 volume=.15).play(fadeout=1000, panning=(random(), random()))

    def cancel_explosion(self):
        self.parent.active_notes.remove(self)
        if self.is_head:
            self.parent.power_ups.supply()


class PowerUps(GameChild):

    BORDER_FILE_NAME = "border.png"
    EFFECT_LENGTH = {"eat enemy": 18000, "grow faster": 14000,
                     "extra heart": 4000, "invincibility": 15000,
                     "slow down": 15000, "spawn enemy": 4000, "speed up": 15000}

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        directory = self.get_resource("power-up")
        self.supply_sound_effects = (\
            SoundEffect(self.get_resource(join(directory, "spotted.wav")), .6),
            SoundEffect(self.get_resource(join(directory, "poison.wav")), .6))
        border_base_image = load(join(directory,
                                      self.BORDER_FILE_NAME)).convert_alpha()
        power_ups = self.power_ups = []
        for path in iglob(join(directory, "*.png")):
            file_name = basename(path)
            effect = self.parse_effect_name(path)
            if file_name != self.BORDER_FILE_NAME:
                power_ups.append(PowerUp(self, load(path).convert_alpha(),
                                         effect, border_base_image.copy(),
                                         self.EFFECT_LENGTH[effect]))
        music = self.music = {}
        for path in iglob(join(directory, "*.ogg")):
            name = self.parse_effect_name(path)
            music[name] = Sound(self.get_resource(path))
            music[name].set_volume(.6)

    def parse_effect_name(self, path):
        return basename(path).replace("-", " ").split(".")[0]

    def reset(self):
        for audio in self.music.itervalues():
            audio.fadeout(1000)
        self.deactivate_power_up()
        for power_up in self.power_ups:
            power_up.reset()

    def deactivate_power_up(self):
        self.active_power_up = None

    def supply(self):
        power_up = self.active_power_up = choice(self.power_ups)
        power_up.activate()
        self.supply_sound_effects[power_up.classification].play()

    def update(self):
        if self.active_power_up:
            self.active_power_up.update()


class PowerUp(Sprite):

    BORDER_COLORS_HSV = ((62, 76, 100), (172, 76, 100)), ((135, 100, 46),
                                                          (246, 100, 46))
    BACKGROUND_BASE_COLORS = Color(0, 255, 0), Color(255, 0, 0)
    BLINK_INTERVAL = 300
    MARGIN = 200
    POSITIVE, NEGATIVE = range(2)
    FRAME_COUNT = 7

    def __init__(self, parent, symbol, effect, border, length):
        Sprite.__init__(self, parent)
        self.symbol = symbol
        self.effect = effect
        self.border = border
        self.length = length
        self.display_surface = self.get_display_surface()
        self.classification = self.NEGATIVE if effect in \
                              ("spawn enemy", "speed up") else self.POSITIVE
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.register(self.cancel_blink, self.highlight, self.restore_speed,
                      self.cancel_growth_increase, self.fade_music_out,
                      self.restore_game_audio, self.restore_enemy,
                      self.deactivate)
        self.set_frames()

    # financial reports rewound reveal
    # vast expenditures of priceless resource allocation
    # feed addiction at the source
    # the NSA's database
    # scan drones
    # and border patrol telecop corporations
    # they say peace, peace is gone

    def set_frames(self):
        self.add_frameset(name="large")
        borders = self.border, self.border.copy()
        pixels = PixelArray(borders[0])
        pixels.replace((0, 0, 0), (15, 15, 15))
        pixels.replace((255, 255, 255), (159, 159, 159))
        pixels = PixelArray(borders[1])
        pixels.replace((255, 255, 255), (255, 0, 0) if \
                       self.classification == self.NEGATIVE else (0, 255, 0))
        pixels.replace((0, 0, 0), (63, 0, 0) if \
                       self.classification == self.POSITIVE else \
                       (0, 63, 0))
        del pixels
        for ii, hue in enumerate(xrange(0, 360, 360 / self.FRAME_COUNT)):
            frame = Surface(borders[0].get_size(), SRCALPHA)
            ratio = float(ii) / (self.FRAME_COUNT - 1)
            color = Color(0, 0, 0)
            color.hsla = hue, 100, 50 , 100
            frame.fill(color)
            frame.blit(borders[bool(ii)], (0, 0))
            color = Color(0, 0, 0)
            color.hsla = (hue + 180) % 360, 100, 50, 100
            symbol = self.symbol.copy()
            pixels = PixelArray(symbol)
            pixels.replace((0, 0, 0), color)
            del pixels
            rect = symbol.get_rect()
            rect.center = frame.get_rect().center
            frame.blit(symbol, rect)
            self.add_frame(frame)

    def act(self):
        # the CIA cloned my sperm to make Obama's third testicle
        length = self.length
        effect = self.effect
        game = self.get_game()
        if effect == "eat enemy":
            if game.enemies.enemies:
                enemy = choice(game.enemies.enemies)
                enemy.set_frameset("edible")
                self.play(self.restore_enemy, delay=length, play_once=True,
                          enemy=enemy)
            else:
                self.halt(self.deactivate)
                self.play(self.deactivate, delay=2000, play_once=True)
        elif effect == "grow faster":
            game.net.set_growth_rate(True)
            self.play(self.cancel_growth_increase, delay=length, play_once=True)
        elif effect == "extra heart":
            game.life_meter.add()
        elif effect == "invincibility":
            game.ships.set_invincible(length)
        elif effect == "slow down":
            game.wires.set_speed(Wires.SLOW)
            game.enemies.set_speed(0)
            self.play(self.restore_speed, delay=length, play_once=True)
        elif effect == "spawn enemy":
            game.enemies.spawn()
        elif effect == "speed up":
            game.wires.set_speed(Wires.FAST)
            game.enemies.set_speed(2)
            self.play(self.restore_speed, delay=length, play_once=True)
        if effect in self.parent.music:
            game.game_audio.fadeout(2000)
            song = self.parent.music[effect]
            song.play(-1, 0, 2000)
            self.play(self.fade_music_out, delay=length - 1000, play_once=True)
            self.play(self.restore_game_audio, delay=length - 1000,
                      play_once=True)

    def cancel_growth_increase(self):
        self.get_game().net.set_growth_rate()

    def restore_speed(self):
        game = self.get_game()
        game.wires.set_speed()
        game.enemies.set_speed(1)

    def fade_music_out(self):
        self.parent.music[self.effect].fadeout(1000)

    def restore_game_audio(self):
        self.get_game().game_audio.play(-1, 0, 1000)

    def restore_enemy(self, enemy):
        enemy.set_frameset(0)

    def deactivate(self):
        self.parent.deactivate_power_up()

    def cancel_eat_enemy(self):
        self.halt(self.restore_enemy)
        self.halt(self.fade_music_out)
        self.halt(self.restore_game_audio)
        self.halt(self.deactivate)
        self.fade_music_out()
        self.restore_game_audio()
        self.deactivate()

    def reset(self):
        w, h = self.location.size
        self.halt(self.blink)
        self.halt(self.cancel_blink)
        self.halt(self.highlight)
        self.halt(self.deactivate)
        self.deactivate()
        self.unhide()
        self.set_alpha(255)

    def blink(self):
        self.toggle_hidden()

    def cancel_blink(self):
        self.halt(self.blink)
        self.hide()

    def highlight(self):
        self.location.midtop = self.display_surface.get_rect().\
                               centerx, Timer.SIZE[1] + 10

    def activate(self):
        self.location.center = self.parent.parent.hands[0].location.center
        self.act()
        self.play(self.blink)
        self.play(self.deactivate, delay=self.length, play_once=True)


class City(GameChild):

    TRANSPARENT_COLOR = 255, 0, 255

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.set_skyline()
        self.lights = Lights(self)

    def set_skyline(self):
        skyline = self.skyline = load(self.get_resource("skyline.png")).\
                  convert()
        skyline.set_colorkey(self.TRANSPARENT_COLOR)
        rect = self.skyline_rect = skyline.get_rect()
        rect.midbottom = self.display_surface.get_rect().midbottom

    def update(self):
        self.display_surface.blit(self.skyline, self.skyline_rect)
        self.lights.update()


class Lights(Animation):

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.set_lights()
        self.register(self.select, interval=400)
        self.play(self.select)

    def set_lights(self):
        surface = load(self.get_resource("lights.png")).convert()
        surface.set_colorkey((0, 0, 0))
        lights = self.lights = []
        x = 0
        y_offset = self.display_surface.get_height() - surface.get_height()
        while x < surface.get_width() - 1:
            y = 0
            while y < surface.get_height():
                it, other = surface.get_at((x, y)), surface.get_at((x + 1, y))
                if it != (0, 0, 0):
                    if it != other and other != (0, 0, 0):
                        tl = x, y - 1
                        lights.append((surface.subsurface(Rect(tl,
                                                               (3, 4))).copy(),
                                       (tl[0], tl[1] + y_offset)))
                    y += 3
                y += 1
            x += 1
        self.select()

    def select(self):
        self.active = set(randrange(0, len(self.lights)) for ii in xrange(20))

    def update(self):
        Animation.update(self)
        for index in self.active:
            self.display_surface.blit(*self.lights[index])


class EndMessage(Sprite):

    def __init__(self, parent, text, hue):
        Sprite.__init__(self, parent)
        self.font = Font(self.get_resource("display", "end-message-font"), 32)
        spaced = text[0]
        for char in text[1:]:
            spaced += "   " + char
        self.set_frames(spaced, hue)
        self.location.center = self.get_display_surface().get_rect().center

    def reset(self):
        self.deactivate()

    def deactivate(self):
        self.active = False

    def activate(self):
        self.active = True

    def set_frames(self, text, hue):
        color = Color(0, 0, 0)
        for lightness in xrange(50, 100, 5):
            color.hsla = hue, 100, lightness, 100
            self.add_frame(self.font.render(text, True, color))

    def update(self):
        if self.active:
            Sprite.update(self)


class Instructions(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        image = load(self.get_resource("instructions.png")).convert_alpha()
        foreground = self.foreground = Surface(image.get_size(), SRCALPHA)
        fr = self.foreground_rect = foreground.get_rect()
        fr.center = ds.get_rect().center
        color = Color(0, 0, 0)
        # color.hsla = 350, 60, 92, 100
        color.hsva = 87, 57, 89, 100
        foreground.fill(color)
        foreground.blit(image, (0, 0))
        shadow = self.shadow = Surface(foreground.get_size())
        # color.hsla = 350, 60, 26, 100
        color.hsva = 345, 57, 25, 100
        shadow.fill(color)
        sr = self.shadow_rect = shadow.get_rect()
        sr.center = fr.centerx - 8, fr.centery + 8

    def reset(self):
        self.activate()

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def update(self):
        if self.active:
            self.display_surface.blit(self.shadow, self.shadow_rect)
            self.display_surface.blit(self.foreground, self.foreground_rect)


class HighScores(Animation):

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.register(self.blink, interval=500)
        self.refresh()
        self.play(self.blink)

    def reset(self):
        self.activate()
        self.blink_on = True

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def refresh(self, submitted=None):
        scores_path = join(self.get_resource("resource"), "scores")
        if not exists(scores_path):
            scores = [0.4025, 0.255, 0.1875, 0.1325, 0.0825]
            open(scores_path, "w").write("\n".join(map(str, scores)) + "\n")
        else:
            scores = map(float, sorted(open(scores_path).readlines(),
                                       reverse=True)[:5])
        font_path = self.get_resource("display", "high-scores-font")
        font = Font(font_path, 14)
        font.set_italic(True)
        font.set_bold(True)
        foreground_color = (192, 127, 40)
        background_color = (254, 220, 178)
        plates = self.plates = [font.render(" HIGH", True, foreground_color,
                                            background_color)]
        self.cancel_blink()
        for ii, score in enumerate(scores):
            size = range(22, 12, -2)[ii]
            if str(score) == submitted:
                self.blink_index = ii
            text = "%.2f" % (score * 100)
            font = Font(font_path, size)
            font.set_italic(True)
            main = font.render(text[:2], True, foreground_color)
            precision = Font(font_path, size - 6).render(text[-3:], True,
                                                    foreground_color)
            background = Surface((main.get_width() + precision.get_width() + 1,
                                  main.get_height()))
            background.fill(background_color)
            background.blit(main, (2, 1))
            rect = precision.get_rect()
            br = background.get_rect()
            rect.bottomright = br.right - 2, br.bottom
            background.blit(precision, rect)
            plates.append(background)

    def cancel_blink(self):
        self.blink_index = None

    def blink(self):
        self.blink_on = not self.blink_on

    def update(self):
        if self.active:
            step = self.display_surface.get_height() / (len(self.plates) + 1)
            for ii, y in enumerate(xrange(step,
                                          self.display_surface.get_height() - \
                                          step, step)):
                plate = self.plates[ii]
                rect = plate.get_rect()
                rect.midleft = 0, y
                if self.blink_on or ii - 1 != self.blink_index:
                    self.display_surface.blit(plate, rect)
            Animation.update(self)
